

#include <iostream.h>
#include <string.h>

#include "CMatch.h"
#include "CNode.h"


//constructor
CMatch::CMatch()
{
	//best and top nodes pointer to nothing
	best = NULL;
	top = NULL;
}


//destructor
CMatch::~CMatch()
{
	//delete top node if necessary
	if (top)
		delete top;
}


//return number of times byte at position is repeated
int CMatch::GetRepeatCount(int position)
{
unsigned char current_value;
int repeat_count;

	//can't check previous value if at beginning of file
	if (position == 0)
		return 0;

	//no repeat sequence found yet
	repeat_count = 0;

	//byte from previous position
	current_value = data[position-1];
	
	//increase repeat count of data keeps the same as current_value
	while ( (current_value == data[position++]) && (position < length) )
		repeat_count++;

	//return found repeat count
	return repeat_count;
}


//return match-length of current data in past data
int CMatch::GetMatchLength(int current, int past)
{
int match_length;	
int current_start = current;	//remember current position to prevent string overlaps

	//no matches possible before start of data...
	if (past < 0)
		return 0;

	//start with no match
	match_length = 0;

	//keep comparing strings while (note that strings may be overlapping for extra compression!)
	while( (current < length) &&				//string doesn't get out of bounds
		   (data[current++] == data[past++]) )	//data is the same
	{
		match_length++;	//increase match length 
	}

	//return found match length
	return match_length;
}


//return number of bits needed to encode value in Elias Gamma Code
//0		: 0
//1/2	: 10x
//3-6	: 110xx
//7-14	: 1110xxx
int CMatch::GetGammaSize(int value)
{
int gamma_size = 1;

	//increase size if there's still bits left after shifting one bit out
	while (value)
	{
		value--;

		gamma_size+=2;	//each time 2 extra bits are needed

		value>>=1;
	}

	//return calculated gamma size
	return gamma_size;
}



void CMatch::GetNextMatch(CNode *parent)
{
CNode *node; 
int cost = parent->GetCost();
int position = parent->GetPosition();
int match_length, match_size;
int best_match_offset = 0;
int best_match_length = 0;
int best_match_size = 1024;
int repeat_count, repeat_size;

	//next level
	depth++;

	//stop compressing if depth is too high or end of file has been reached
	if ( (depth > 8) || (position == length) )
	{	//if no best node found yet
		if (best->GetType() == NODE_TYPE_TOP)
		{
			best->Init(parent, NODE_TYPE_OPTIMAL, parent->GetCost(), parent->GetPosition());
		}
		else
		{	//best node has already been defined

			//check if new is actually better than current best node
			if (parent->GetPosition() > best->GetPosition() )
			{	//if so, change best node
				best->Init(parent, NODE_TYPE_OPTIMAL, parent->GetCost(), parent->GetPosition());
			}
			else
			{	//maybe this node costs less than best node?
				if ( (parent->GetPosition() == best->GetPosition()) &&
					 (parent->GetCost() < best->GetCost()) )
				{	//change best node
					best->Init(parent, NODE_TYPE_OPTIMAL, parent->GetCost(), parent->GetPosition());
				}
			}
		}

		//previous level
		depth--;

		return;
	}

	//create new literal node
	node = new CNode();

	//init literal, costs 9 bits, advances 1 byte in data
	node->Init(parent, NODE_TYPE_LITERAL, cost + 9, position + 1);

	node->InitLiteral(data[position]);

	//add literal node to parent
	parent->SetLiteral(node);

	//start compression from new literal node
	GetNextMatch(node);

	//find best string match
	for (int match_offset = 2; match_offset < 2047; match_offset++)
	{
		//get match length from this position
		match_length = GetMatchLength(parent->GetPosition(), parent->GetPosition() - match_offset);

		//compression possible of length > 2
		if (match_length > 1)
		{
			//if (match_offset == match_length)
			//	match_length = GetMatchExtend(parent->GetPosition(), match_length);

			//limit maximum length
			if (match_length > 65535)
				match_length = 65535;

			//new best match if current match is better than best match
			if ( match_length > best_match_length )
			{
				//use 8 or 12 bits for encoding offset
				if (match_offset < 128 + 1)
					match_size = GetGammaSize(match_length - 2) + 1 + 8;
				else
					match_size = GetGammaSize(match_length - 2) + 1 + 12;

				//change best match found
				best_match_offset = match_offset;
				best_match_length = match_length;
				best_match_size = match_size;
			}
		}
	}

	//found a match if length is at least 2
	if (best_match_length > 1)
	{
		//create new match node
		node = new CNode();
		
		//initialise new match node
		node->Init(parent, NODE_TYPE_MATCH, cost + best_match_size, position + best_match_length);

		node->InitMatch(best_match_offset, best_match_length);

		//add match node to parent
		parent->SetMatch(node);

		//continue compressing from new point
		GetNextMatch(node);
	}

	//get number of times current byte is repeated
	repeat_count = GetRepeatCount(parent->GetPosition());

	//we need at least 2 equal bytes for compression
	if ( repeat_count > 1 )
	{
		//limit maximum length
		if (repeat_count > 65535)
			repeat_count = 65535;

		//calculate number of bits to encode this repetition
		repeat_size = GetGammaSize(repeat_count - 2) + 1 + 8;

		//create new repeat node
		node = new CNode();

		//initialise new repeat node
		node->Init(parent, NODE_TYPE_REPEAT, cost + repeat_size, position + repeat_count);

		node->InitRepeat(repeat_count);

		//add repeat node to parent
		parent->SetRepeat(node);

		//continue compression after repeated data
		GetNextMatch(node);
	}

	//previous level
	depth--;
}


//init data
void CMatch::Init(unsigned char *data, int length)
{
	this->data = data;
	this->length = length;
}


//find best match from starting point
CNode* CMatch::FindMatches(int position)
{
	//delete old top node
	if (top)
		delete top;

	//create new top node
	top = new CNode();

	//create new best node
	best = new CNode();

	//initialise top node
	top->Init(NULL, NODE_TYPE_TOP, 0, position);

	//initialise best node
	best->Init(NULL, NODE_TYPE_OPTIMAL, 0, 0);

	//current recursion depth is 0
	depth = 0;

	//find next match
	GetNextMatch(top);

	//return best match found
	return best;
}
